unit AGFUnit;
//---------------------------------------------------------------------------
// TAGFImage v3.0 (c) 2000 - 2003
// Written by Lifepower (lifepower@fromru.com)
//---------------------------------------------------------------------------
interface

//---------------------------------------------------------------------------
uses
 Windows, SysUtils, Classes, d3d9, PowerTypes, VTDbUnit;

//---------------------------------------------------------------------------
type
 // for pixel processing routines
 TPixelFunc = function(Color: Cardinal; Tag: Integer): Cardinal of object;
//---------------------------------------------------------------------------
 TAGFImage = class
 private
  FInitialized: Boolean;

  PowerD3D: TPowerD3D;

  pWidth,
  pHeight,
  tWidth,
  tHeight,
  pCount,
  ImgPerW,
  ImgPerH,
  ImgPerTex: Integer;

  FFormat: TD3DFormat;

  FTransColor: Cardinal;

  Textures: array of IDirect3DTexture9;
  function CountTextures(): Integer;

  function GetTexture(Num: Integer): IDirect3DTexture9;
  function GetBytesPerPixel(): Integer;

  function GetTextureData(var Write: Pointer; TexNum: Integer): Integer;
  function SetTextureData(var Read: Pointer; TexNum: Integer): Integer;

  function GetPixel(X, Y, TexNum: Integer): Cardinal;
  procedure SetPixel(X, Y, TexNum: Integer; Value: Cardinal);
  function GetPixelAlpha(X, Y, TexNum: Integer): Integer;
  procedure SetPixelAlpha(X, Y, TexNum, Value: Integer);
  procedure SetTransColor(Value: Cardinal);
  function CheckFormat(Format: TD3DFormat): Boolean;
  function negColorFunc(Color: Cardinal; Tag: Integer): Cardinal;
 public
  property Initialized: Boolean read FInitialized;
  property PatternWidth: Integer read PWidth;
  property PatternHeight: Integer read PHeight;
  property PatternCount: Integer read PCount;

  property TextureWidth: Integer read TWidth;
  property TextureHeight: Integer read THeight;
  property TextureCount: Integer read CountTextures;

  property Texture[Num: Integer]: IDirect3DTexture9 read GetTexture;
  property Format: TD3DFormat read FFormat;
  property BytesPerPixel: Integer read GetBytesPerPixel;

  property ImagesPerTexture: Integer read ImgPerTex;
  property ImagesPerWidth: Integer read ImgPerW;
  property ImagesPerHeight: Integer read ImgPerH;

  property TransparentColor: Cardinal read FTransColor write SetTransColor;

  property FormatSupported[Format: TD3DFormat]: Boolean read CheckFormat;

  // use to set/get 32-bit pixel value on specific texture (TexNum: [0..TextureCount -1])
  property Pixels[X, Y, TexNum: Integer]: Cardinal read GetPixel write SetPixel;

  // use to set/get alpha value on specific texture (TexNum: [0..TextureCount -1])
  property PixelsAlpha[X, Y, TexNum: Integer]: Integer read GetPixelAlpha write SetPixelAlpha;


  //---------------------------------------------------------------------------
  // - FOR INTERNAL USE
  // returns IDirect3DTexture9 based on source coordinate parameters
  function AcquireTexture(TexCoord: TTexCoord; var VertexArray: TVertexArray4;
   out Texture: IDirect3DTexture9): Integer;


  constructor Create(_PowerD3D: TPowerD3D);

  function Initialize(TexWidth, TexHeight, TexCount, PtWidth, PtHeight: Integer;
   TexFormat: TD3DFormat): Integer;

  function Finalize(): Integer;

  // lock/unlock the texture to gain access to its pixel data
  function Lock(TexNum: Integer; out LRect: TD3DLocked_Rect): Integer;
  function Unlock(TexNum: Integer): Integer;

  // returns usable AGF texture data
  function GetAGFData(out Data: Pointer; out DataSize: Integer): Integer;

  // sets AGF texture data
  function SetAGFData(Data: Pointer; DataSize: Integer; DFormat: TD3DFormat): Integer;

  // loads the image from VTDb-compatible file
  function LoadFromVTDb(VTDb: TVTDb; Key: TRecordKey; TexFormat: TD3DFormat): Integer;

  // saves the image to VTDb-compatible file
  function SaveToVTDb(VTDb: TVTDb; Key: TRecordKey): Integer;

  // loads textures from stream and converts them to the desired format
  function ReadFromStream(Stream: TStream; TexFormat: TD3DFormat): Integer;

  // saves textures in 32-bit format to stream
  function WriteToStream(Stream: TStream): Integer;

  // this method loads the texture from BMP or TGA file
  function LoadFromFile(FileName: string; PatternWidth, PatternHeight, TextureWidth,
   TextureHeight: Integer; TextureFormat: TD3DFormat): Integer;

  function LoadFromStream(M: Tmemorystream; PatternWidth, PatternHeight, TextureWidth, TextureHeight: Integer; TextureFormat: TD3DFormat): Integer;
  //---------------------------------------------------------------------------
  // image processing effects
  //---------------------------------------------------------------------------
  // calls func for every pixel in AGF image
  function imgProcess(func: TPixelFunc; Tag: Integer): Integer;

  // makes either negative image, negative alpha or both
  function imgNegative(Color, Alpha: Boolean): Integer;

  // converts image to grayscale
  function imgGrayscale(): Integer;

  // either mirrors or flips the image (BUGGY: only Flips/Mirrors texture, not actual patterns!)
  function imgInvert(Mirror, Flip: Boolean): Integer;
 end;

//---------------------------------------------------------------------------
implementation
//---------------------------------------------------------------------------
uses
 Graphics, PStHandle, {$IFDEF PDrawEx2}PDrawEx2{$ELSE}PDrawEx{$ENDIF}, TGAReader,
 Math;

//---------------------------------------------------------------------------
constructor TAGFImage.Create(_PowerD3D: TPowerD3D);
begin
 inherited Create();

 PWidth:= 0;
 PHeight:= 0;
 TWidth:= 0;
 THeight:= 0;
 PCount:= 0;
 ImgPerW:= 0;
 ImgPerH:= 0;
 ImgPerTex:= 0;
 FFormat:= D3DFMT_UNKNOWN;

 FTransColor:= 0;

 SetLength(Textures, 0);

 PowerD3D:= _PowerD3D;

 FInitialized:= False;
end;

//---------------------------------------------------------------------------
function TAGFImage.Initialize(TexWidth, TexHeight, TexCount, PtWidth, PtHeight: Integer;
 TexFormat: TD3DFormat): Integer;
var
 i, Res: Integer;
begin
 if (TexCount < 1)or(FInitialized)or(PtWidth > TexWidth)or(PtHeight > TexHeight)or
  (PtWidth < 0)or(PtHeight < 0)or(not PowerD3D.Initialized)or(not PowerD3D.DeviceActive) then
  begin
   Result:= errInvalidCall;
   Exit;
  end;

 // initialize textures
 SetLength(Textures, TexCount);
 for i:= 0 to Length(Textures) - 1 do
  Textures[i]:= nil;

 // create textures
 for i:= 0 to Length(Textures) - 1 do
  begin
   Res:= PowerD3D.D3DDevice9.CreateTexture(TexWidth, TexHeight, 1, 0, TexFormat,
    D3DPOOL_MANAGED, Textures[I], nil);

   if (Res <> D3D_OK) then
    begin
     // finalize all previously created textures
     Finalize();

     Result:= errCreateTexFail;
     Exit;
    end;
  end;

 // texture format
 FFormat:= TexFormat;

 // texture dimensions
 TWidth:= TexWidth;
 THeight:= TexHeight;

 // pattern width & height
 if (PtWidth = 0) then PWidth:= TWidth else PWidth:= PtWidth;
 if (PtHeight = 0) then PHeight:= THeight else PHeight:= PtHeight;

 // how much patterns fit in texture
 ImgPerW:= TWidth div PWidth;
 ImgPerH:= THeight div PHeight;
 ImgPerTex:= ImgPerW * ImgPerH;
 PCount:= ImgPerTex * Length(Textures);

 FInitialized:= True;
 Result:= errNone;
end;

//---------------------------------------------------------------------------
function TAGFImage.Finalize(): Integer;
var
 i: Integer;
begin
 Result:= errNone;

 // finalize textures
 for i:= 0 to Length(Textures) - 1 do
  Textures[i]:= nil;

 SetLength(Textures, 0);
 FInitialized:= False;
end;

//---------------------------------------------------------------------------
function TAGFImage.Lock(TexNum: Integer; out LRect: TD3DLocked_Rect): Integer;
begin
 FillChar(LRect, SizeOf(LRect), 0);

 if (not FInitialized)or(TexNum >= Length(Textures)) then
  begin
   Result:= errInvalidCall;
   Exit;
  end;

 if (Textures[TexNum].LockRect(0, LRect, nil, 0) <> D3D_OK) then
  begin
   Result:= errDeviceError;
   Exit;
  end;

 Result:= errNone;
end;

//---------------------------------------------------------------------------
function TAGFImage.Unlock(TexNum: Integer): Integer;
begin
 if (not FInitialized)or(TexNum >= Length(Textures)) then
  begin
   Result:= errInvalidCall;
   Exit;
  end;

 if (Textures[TexNum].UnlockRect(0) <> D3D_OK) then
  begin
   Result:= errDeviceError;
   Exit;
  end;

 Result:= errNone;
end;

//---------------------------------------------------------------------------
function TAGFImage.CountTextures(): Integer;
begin
 Result:= Length(Textures);
end;

//---------------------------------------------------------------------------
function TAGFImage.GetTexture(Num: Integer): IDirect3DTexture9;
begin
 if (Num < 0)or(Num >= Length(Textures)) then
  begin
   Result:= nil;
   Exit;
  end;

 Result:= Textures[Num];
end;

//---------------------------------------------------------------------------
function TAGFImage.GetBytesPerPixel(): Integer;
begin
 case Format of
  D3DFMT_R3G3B2: Result:= 1;
  D3DFMT_R5G6B5: Result:= 2;
  D3DFMT_X8R8G8B8: Result:= 4;
  D3DFMT_X1R5G5B5: Result:= 2;
  D3DFMT_X4R4G4B4: Result:= 2;
  D3DFMT_A8R8G8B8: Result:= 4;
  D3DFMT_A1R5G5B5: Result:= 2;
  D3DFMT_A4R4G4B4: Result:= 2;
  D3DFMT_A8R3G3B2: Result:= 2;

  else Result:= 0;
 end;
end;

//---------------------------------------------------------------------------
function TAGFImage.CheckFormat(Format: TD3DFormat): Boolean;
begin
 Result:= (Format in [D3DFMT_R3G3B2, D3DFMT_R5G6B5, D3DFMT_X8R8G8B8, D3DFMT_X1R5G5B5,
  D3DFMT_X4R4G4B4, D3DFMT_A8R8G8B8, D3DFMT_A1R5G5B5, D3DFMT_A4R4G4B4, D3DFMT_A8R3G3B2]);
end;

//---------------------------------------------------------------------------
function TAGFImage.SetTextureData(var Read: Pointer; TexNum: Integer): Integer;
var
 LRect: TD3DLocked_Rect;
 i: Integer;
 Dest: Pointer;
begin
 if (not FInitialized) then
  begin
   Result:= errInvalidCall;
   Exit;
  end;

 // load PDrawEx.DLL
 if (not PDrawExReady)and(not InitializePDrawEx()) then
  begin
   Result:= errPDrawExFail;
   Exit;
  end;

 // lock the texture
 Result:= Lock(TexNum, LRect);
 if (Result <> 0) then Exit;

 // convert pixel data 
 for i:= 0 to THeight - 1 do
  begin
   Integer(Dest):= Integer(LRect.pBits) + (i * LRect.Pitch);

   if (pdrawLineConv(Read, Dest, TWidth, D3DFMT_A8R8G8B8, FFormat) <> 0) then
    begin
     Result:= errInvalidFormat;
     Exit;
    end;

   Inc(Integer(Read), TWidth * 4);
  end;

 // unlock the texture
 Result:= Unlock(TexNum);
end;

//---------------------------------------------------------------------------
function TAGFImage.GetTextureData(var Write: Pointer; TexNum: Integer): Integer;
var
 LRect: TD3DLocked_Rect;
 i: Integer;
 Source: Pointer;
begin
 if (not FInitialized) then
  begin
   Result:= errInvalidCall;
   Exit;
  end;

 // load PDrawEx.dll
 if (not PDrawExReady)and(not InitializePDrawEx()) then
  begin
   Result:= errPDrawExFail;
   Exit;
  end;

 // lock the texture
 Result:= Lock(TexNum, LRect);
 if (Result <> 0) then Exit;

 // convert pixel data
 for i:= 0 to THeight - 1 do
  begin
   Integer(Source):= Integer(LRect.pBits) + (i * LRect.Pitch);

   if (pdrawLineConv(Source, Write, TWidth, FFormat, D3DFMT_A8R8G8B8) <> 0) then
    begin
     Result:= errInvalidFormat;
     Exit;
    end;

   Inc(Integer(Write), TWidth * 4);
  end;

 // unlock the texture
 Result:= Unlock(TexNum);
end;

//---------------------------------------------------------------------------
function TAGFImage.SetAGFData(Data: Pointer; DataSize: Integer; DFormat: TD3DFormat): Integer;
var
 Read: Pointer;
 i, TexWidth, TexHeight, PtWidth, PtHeight, TexCount, ID: Integer;
begin
 if (FInitialized) then
  begin
   Result:= errInvalidCall;
   Exit;
  end;

 if (not CheckFormat(DFormat)) then
  begin
   Result:= errInvalidFormat;
   Exit;
  end;

 Read:= Data;
 // read data ID
 ID:= ReadInt(Read);
 if (ID <> $58464741) then
  begin
   Result:= errInvalidData;
   Exit;
  end;

 // texture dimensions
 TexWidth:= ReadInt(Read);
 TexHeight:= ReadInt(Read);
 // pattern sizes
 PtWidth:= ReadInt(Read);
 PtHeight:= ReadInt(Read);
 // texture count
 TexCount:= ReadInt(Read);

 // initialize texture
 Result:= Initialize(TexWidth, TexHeight, TexCount, PtWidth, PtHeight, DFormat);
 if (Result <> 0) then Exit;

 // load textures
 for i:= 0 to Length(Textures) - 1 do
  begin
   Result:= SetTextureData(Read, i);
   if (Result <> 0) then
    begin
     Finalize();
     Exit;
    end;
  end;

 Result:= errNone;
end;

//---------------------------------------------------------------------------
function TAGFImage.GetAGFData(out Data: Pointer; out DataSize: Integer): Integer;
var
 Write: Pointer;
 i: Integer;
begin
 if (not FInitialized)or(Length(Textures) < 1) then
  begin
   Result:= errInvalidCall;
   Exit;
  end;

 DataSize:= 24 + Length(Textures) * TWidth * THeight * 4;
 Data:= AllocMem(DataSize);
 Write:= Data;

 // agf ID
 AddInt(Write, $58464741);

 // texture dimensions
 AddInt(Write, TWidth);
 AddInt(Write, THeight);
 // pattern sizes
 AddInt(Write, PWidth);
 AddInt(Write, PHeight);
 // texture count
 AddInt(Write, Length(Textures));

 // process textures
 for i:= 0 to Length(Textures) - 1 do
  begin
   Result:= GetTextureData(Write, i);
   if (Result <> 0) then
    begin
     FreeMem(Data);
     Data:= nil;
     Exit;
    end;
  end;

 Result:= errNone;
end;

//---------------------------------------------------------------------------
function TAGFImage.LoadFromVTDb(VTDb: TVTDb; Key: TRecordKey; TexFormat: TD3DFormat): Integer;
var
 Source: Pointer;
 SourceSize, rType: Integer;
begin
 if (FInitialized) then
  begin
   Result:= errInvalidCall;
   Exit;
  end;

 // read VTDb record
 Result:= VTDb.ReadRecordEx(Key, Source, SourceSize, rType);
 if (Result <> errNone) then Exit;

 // initialize textures
 if (VTDb.Version >= 1.1)and(rType <> recAGFImage)and(UpperCase(Key) <> 'FONT') then
  begin
   Result:= errNotAGF;
   Exit;
  end;

 // initialize textures
 Result:= SetAGFData(Source, SourceSize, TexFormat);

 FreeMem(Source);
end;

//---------------------------------------------------------------------------
function TAGFImage.SaveToVTDb(VTDb: TVTDb; Key: TRecordKey): Integer;
var
 Data: Pointer;
 DataSize: Integer;
begin
 if (FInitialized) then
  begin
   Result:= errInvalidCall;
   Exit;
  end;

 // read AGF data
 Result:= GetAGFData(Data, DataSize);
 if (Result <> errNone) then Exit;

 // write VTDb record
 Result:= VTDb.WriteRecord(Key, Data, DataSize, recAGFImage);
 if (Result <> errNone) then Exit;

 FreeMem(Data);
end;

//---------------------------------------------------------------------------
function TAGFImage.ReadFromStream(Stream: TStream; TexFormat: TD3DFormat): Integer;
var
 Source: Pointer;
 SourceSize: Integer;
begin
 if (FInitialized) then
  begin
   Result:= errInvalidCall;
   Exit;
  end;

 // read data size
 Stream.ReadBuffer(SourceSize, SizeOf(Integer));
 // allocate memory for buffer
 GetMem(Source, SourceSize);
 // read AGF data
 Stream.ReadBuffer(Source^, SourceSize);
 // initialize textures
 Result:= SetAGFData(Source, SourceSize, TexFormat);
 // release buffer
 FreeMem(Source);
end;

//---------------------------------------------------------------------------
function TAGFImage.WriteToStream(Stream: TStream): Integer;
var
 Data: Pointer;
 DataSize: Integer;
begin
 if (FInitialized) then
  begin
   Result:= errInvalidCall;
   Exit;
  end;

 // get writeable AGF data
 Result:= GetAGFData(Data, DataSize);
 if (Result <> errNone) then Exit;

 // write data size
 Stream.WriteBuffer(DataSize, SizeOf(Integer));
 Stream.WriteBuffer(Data^, DataSize);

 FreeMem(Data);
end;

//---------------------------------------------------------------------------
function TAGFImage.AcquireTexture(TexCoord: TTexCoord; var VertexArray: TVertexArray4;
 out Texture: IDirect3DTexture9): Integer;
var
 TexNum, i: Integer;
 FloatX1, FloatY1,
 FloatX2, FloatY2,
 SrcX, SrcY: Single;
begin
 Texture:= nil;

 if (not FInitialized)or(Length(Textures) < 1)or(PCount < 1) then
  begin
   Result:= errInvalidCall;
   Exit;
  end;

 // clip pattern number to available patterns
 if (TexCoord.Pattern < 0) then TexCoord.Pattern:= 0;
 if (TexCoord.Pattern > PCount - 1) then TexCoord.Pattern:= PCount - 1;

 // select the specified texture
 TexNum:= TexCoord.Pattern div ImgPerTex;
 Texture:= Textures[TexNum];

 // calculate vertex coordinates
 i:= TexCoord.Pattern mod ImgPerTex;
 SrcX:= (i mod ImgPerW) * PWidth;
 SrcY:= ((i div ImgPerW) mod ImgPerH) * PHeight;
 FloatX1:= (SrcX + TexCoord.SrcX) / TWidth;
 FloatY1:= (SrcY + TexCoord.SrcX) / THeight;
 if (TexCoord.Width > 0) then
  FloatX2:= ((SrcX + TexCoord.Width) / TWidth)
   else FloatX2:= ((SrcX + PWidth) / TWidth);

 if (TexCoord.Height > 0) then
  FloatY2:= ((SrcY + TexCoord.Height) / THeight)
   else FloatY2:= ((SrcY + PHeight) / THeight);

 if (TexCoord.Mirror) then
  begin
   VertexArray[0].tu:= FloatX2;
   VertexArray[1].tu:= FloatX1;
   VertexArray[2].tu:= FloatX2;
   VertexArray[3].tu:= FloatX1;
  end else
  begin
   VertexArray[0].tu:= FloatX1;
   VertexArray[1].tu:= FloatX2;
   VertexArray[2].tu:= FloatX1;
   VertexArray[3].tu:= FloatX2;
  end;

 if (TexCoord.Flip) then
  begin
   VertexArray[0].tv:= FloatY2;
   VertexArray[1].tv:= FloatY2;
   VertexArray[2].tv:= FloatY1;
   VertexArray[3].tv:= FloatY1;
  end else
  begin
   VertexArray[0].tv:= FloatY1;
   VertexArray[1].tv:= FloatY1;
   VertexArray[2].tv:= FloatY2;
   VertexArray[3].tv:= FloatY2;
  end;

 Result:= errNone;
end;

//---------------------------------------------------------------------------
function TAGFImage.GetPixel(X, Y, TexNum: Integer): Cardinal;
var
 LRect: TD3DLocked_Rect;
 SrcPtr: Pointer;
begin
 Result:= 0;
 if (not FInitialized) then Exit;
 if (Lock(TexNum, LRect) <> errNone) then Exit;

 SrcPtr:= Pointer(Integer(LRect.pBits) + (LRect.Pitch * y) + (x * BytesPerPixel));

 pdrawFormatConv(SrcPtr, @Result, FFormat, D3DFMT_A8R8G8B8);

 Unlock(TexNum);
end;

//---------------------------------------------------------------------------
procedure TAGFImage.SetPixel(X, Y, TexNum: Integer; Value: Cardinal);
var
 LRect: TD3DLocked_Rect;
 DestPtr: Pointer;
begin
 if (Lock(TexNum, LRect) <> errNone) then Exit;

 DestPtr:= Pointer(Integer(LRect.pBits) + (LRect.Pitch * y) + (x * BytesPerPixel));

 pdrawFormatConv(@Value, DestPtr, D3DFMT_A8R8G8B8, FFormat);

 Unlock(TexNum);
end;

//---------------------------------------------------------------------------
function TAGFImage.GetPixelAlpha(X, Y, TexNum: Integer): Integer;
var
 LRect: TD3DLocked_Rect;
 SrcPtr: Pointer;
 Color: Cardinal;
begin
 Result:= 0;
 if (not FInitialized) then Exit;
 if (Lock(TexNum, LRect) <> errNone) then Exit;

 SrcPtr:= Pointer(Integer(LRect.pBits) + (LRect.Pitch * y) + (x * BytesPerPixel));

 pdrawFormatConv(SrcPtr, @Color, FFormat, D3DFMT_A8R8G8B8);

 Unlock(TexNum);
 Result:= Integer(Color shr 24);
end;

//---------------------------------------------------------------------------
procedure TAGFImage.SetPixelAlpha(X, Y, TexNum, Value: Integer);
var
 LRect: TD3DLocked_Rect;
 SrcPtr: Pointer;
 Color: Cardinal;
begin
 if (not FInitialized) then Exit;

 if (Lock(TexNum, LRect) <> errNone) then Exit;

 SrcPtr:= Pointer(Integer(LRect.pBits) + (LRect.Pitch * y) + (x * BytesPerPixel));

 // read the pixel first
 pdrawFormatConv(SrcPtr, @Color, FFormat, D3DFMT_A8R8G8B8);
 // apply new alpha
 Color:= (Color and $FFFFFF) or (Cardinal(Value) shl 24);
 // write the altered pixel
 pdrawFormatConv(@Color, SrcPtr, D3DFMT_A8R8G8B8, FFormat);

 Unlock(TexNum);
end;

//---------------------------------------------------------------------------
procedure TAGFImage.SetTransColor(Value: Cardinal);
var
 LRect: TD3DLocked_Rect;
 SrcPtr: Pointer;
 Color: Cardinal;
 k, j, i, Bpp: Integer;
begin
 FTransColor:= Value;
 if (not FInitialized) then Exit;

 Bpp:= BytesPerPixel;
 Value:= Value and $00FFFFFF;

 for k:= 0 to Length(Textures) - 1 do
  begin
   if (Lock(k, LRect) <> errNone) then Exit;

   for j:= 0 to THeight - 1 do
    begin
     SrcPtr:= Pointer(Integer(LRect.pBits) + (LRect.Pitch * j));
     for i:= 0 to TWidth - 1 do
      begin
       // read source pixel
       pdrawFormatConv(SrcPtr, @Color, FFormat, D3DFMT_A8R8G8B8);
       // test for transparency
       if (Color and $00FFFFF = Value) then
        Color:= Color and $00FFFFFF
         else Color:= Color or $FF000000;

       // write the altered pixel
       pdrawFormatConv(@Color, SrcPtr, D3DFMT_A8R8G8B8, FFormat);

       Inc(Integer(SrcPtr), Bpp);
      end;
    end;

   if (Unlock(k) <> errNone) then Exit;
  end;
end;

//---------------------------------------------------------------------------
function TAGFImage.LoadFromFile(FileName: string; PatternWidth, PatternHeight, TextureWidth,
 TextureHeight: Integer; TextureFormat: TD3DFormat): Integer;
var
 Bitmap: TBitmapEx;
 ImgNum: Integer;
 PPL, Bpp,
 i, j, k, TexCount,
 SrcImg, DstImg, SrcX, SrcY, DestX, DestY, PixCount: Integer;
 LRect: TD3DLocked_Rect;
 SrcP, DstP: Pointer;

 SrcFormat: TD3DFormat;
begin
 if (FInitialized) then
  begin
   Result:= errInvalidCall;
   Exit;
  end;
 // initialize PDrawEx.DLL
 if (not PDrawExReady)and(not InitializePDrawEx()) then
  begin
   Result:= errPDrawExFail;
   Exit;
  end;
 // check if the specified format is supported
 if (not CheckFormat(TextureFormat)) then
  begin
   Result:= errInvalidFormat;
   Exit;
  end;

 // load source image
 Bitmap:= TBitmapEx.Create();
 Bitmap.LoadFromFile(FileName);
 // check if source file contains alpha-channel
 if (Bitmap.PixelFormat = pf32bit) then
  SrcFormat:= D3DFMT_A8R8G8B8
   else SrcFormat:= D3DFMT_X8R8G8B8;
 // convert source format to 32-bit format  
 Bitmap.PixelFormat:= pf32Bit;

 // calculate pattern size
 if (PatternWidth = 0) then PatternWidth:= Bitmap.Width;
 if (PatternHeight = 0) then PatternHeight:= Bitmap.Height;

 // check for valid texture and pattern sizes
 if (TextureWidth < PatternWidth)or(TextureHeight < PatternHeight)or
  (PatternWidth > Bitmap.Width)or(PatternHeight > Bitmap.Height) then
  begin
   Result:= errInvalidCall;
   Exit;
  end;

 // set specified parameters
 PWidth:= PatternWidth;
 PHeight:= PatternHeight;
 TWidth:= TextureWidth;
 THeight:= TextureHeight;

 // calculate pattern variables
 PCount:= (Bitmap.Width div PWidth) * (Bitmap.Height div PHeight);
 PPL:= Bitmap.Width div PWidth;
 ImgPerW:= TWidth div PWidth;
 ImgPerH:= THeight div PHeight;
 ImgPerTex:= ImgPerW * ImgPerH;
 TexCount:= Ceil(PCount / ImgPerTex);

 // initialize textures
 Result:= Initialize(TWidth, THeight, TexCount, PWidth, PHeight, TextureFormat);
 if (Result <> errNone) then
  begin
   Bitmap.Free();
   Exit;
  end;

 Bpp:= BytesPerPixel;

 // tile patterns
 SrcImg:= 0;
 for i:= 0 to TexCount - 1 do
  begin
   // destination pattern
   DstImg:= 0;

   // lock the texture
   Result:= Lock(i, LRect);
   if (Result <> errNone) then
    begin
     Bitmap.Free();
     Exit;
    end;

   // tile images
   for j:= 0 to ImgPerTex - 1 do
    begin
     // calculate source and destination coordinates
     DestX:= (DstImg mod ImgPerW) * PWidth;
     DestY:= (DstImg div ImgPerW) * PHeight;
     SrcX:= (SrcImg mod PPL) * PWidth;
     SrcY:= (SrcImg div PPL) * PHeight;

     // write scanlines
     for k:= 0 to PHeight - 1 do
      begin
       // skip remaining vertical lines
       if (SrcY + PHeight > Bitmap.Height) then Break;

       // calculate source image pointer
       Integer(SrcP):= Integer(Bitmap.ScanLine[(k + SrcY)]) + (SrcX * 4);

       // skip remaining horizontal lines
       PixCount:= PWidth;
       if ((SrcX + PixCount) > Bitmap.Width - 1) then
        PixCount:= Bitmap.Width - SrcX;

       // calculate destination pointer
       Integer(DstP):= Integer(LRect.pBits) + ((k + DestY) * LRect.Pitch) + (DestX * Bpp);

       // convert pixels
       pdrawLineConv(SrcP, DstP, PixCount, D3DFMT_A8R8G8B8, TextureFormat);
      end; // for k:= 0 to PHeight - 1 do

     // increment source & destination images
     Inc(SrcImg);
     Inc(DstImg);
    end;

   // unlock the texture
   Result:= Unlock(i);
   if (Result <> errNone) then
    begin
     Bitmap.Free();
     Exit;
    end;
  end;

 // release the bitmap memory
 Bitmap.Free();
 Result:= errNone;
end;

function TAGFImage.LoadFromStream(M: Tmemorystream; PatternWidth, PatternHeight, TextureWidth,
 TextureHeight: Integer; TextureFormat: TD3DFormat): Integer;
var
 Bitmap: TBitmapEx;
 Bmp:Tbitmap;
 ImgNum: Integer;
 PPL, Bpp,
 i, j, k, TexCount,
 SrcImg, DstImg, SrcX, SrcY, DestX, DestY, PixCount: Integer;
 LRect: TD3DLocked_Rect;
 SrcP, DstP: Pointer;
 SrcFormat: TD3DFormat;
begin
if (FInitialized) then
  begin
   Result:= errInvalidCall;
   Exit;
  end;
 // initialize PDrawEx.DLL
 if (not PDrawExReady)and(not InitializePDrawEx()) then
  begin
   Result:= errPDrawExFail;
   Exit;
  end;
 // check if the specified format is supported
 if (not CheckFormat(TextureFormat)) then
  begin
   Result:= errInvalidFormat;
   Exit;
  end;

 // load source image
 Bmp:=Tbitmap.Create;
 bmp.LoadFromStream(m);
 Bitmap:= TBitmapEx.Create();
 bitmap.Assign(bmp);
 bmp.Free;
 // check if source file contains alpha-channel
 if (Bitmap.PixelFormat = pf32bit) then
  SrcFormat:= D3DFMT_A8R8G8B8
   else SrcFormat:= D3DFMT_X8R8G8B8;
 // convert source format to 32-bit format
 Bitmap.PixelFormat:= pf32Bit;

 // calculate pattern size
 if (PatternWidth = 0) then PatternWidth:= Bitmap.Width;
 if (PatternHeight = 0) then PatternHeight:= Bitmap.Height;

 // check for valid texture and pattern sizes
 if (TextureWidth < PatternWidth)or(TextureHeight < PatternHeight)or
  (PatternWidth > Bitmap.Width)or(PatternHeight > Bitmap.Height) then
  begin
   Result:= errInvalidCall;
   Exit;
  end;

 // set specified parameters
 PWidth:= PatternWidth;
 PHeight:= PatternHeight;
 TWidth:= TextureWidth;
 THeight:= TextureHeight;

 // calculate pattern variables
 PCount:= (Bitmap.Width div PWidth) * (Bitmap.Height div PHeight);
 PPL:= Bitmap.Width div PWidth;
 ImgPerW:= TWidth div PWidth;
 ImgPerH:= THeight div PHeight;
 ImgPerTex:= ImgPerW * ImgPerH;
 TexCount:= Ceil(PCount / ImgPerTex);

 // initialize textures
 Result:= Initialize(TWidth, THeight, TexCount, PWidth, PHeight, TextureFormat);
 if (Result <> errNone) then
  begin
   Bitmap.Free();
   Exit;
  end;

 Bpp:= BytesPerPixel;

 // tile patterns
 SrcImg:= 0;
 for i:= 0 to TexCount - 1 do
  begin
   // destination pattern
   DstImg:= 0;

   // lock the texture
   Result:= Lock(i, LRect);
   if (Result <> errNone) then
    begin
     Bitmap.Free();
     Exit;
    end;

   // tile images
   for j:= 0 to ImgPerTex - 1 do
    begin
     // calculate source and destination coordinates
     DestX:= (DstImg mod ImgPerW) * PWidth;
     DestY:= (DstImg div ImgPerW) * PHeight;
     SrcX:= (SrcImg mod PPL) * PWidth;
     SrcY:= (SrcImg div PPL) * PHeight;

     // write scanlines
     for k:= 0 to PHeight - 1 do
      begin
       // skip remaining vertical lines
       if (SrcY + PHeight > Bitmap.Height) then Break;

       // calculate source image pointer
       Integer(SrcP):= Integer(Bitmap.ScanLine[(k + SrcY)]) + (SrcX * 4);

       // skip remaining horizontal lines
       PixCount:= PWidth;
       if ((SrcX + PixCount) > Bitmap.Width - 1) then
        PixCount:= Bitmap.Width - SrcX;

       // calculate destination pointer
       Integer(DstP):= Integer(LRect.pBits) + ((k + DestY) * LRect.Pitch) + (DestX * Bpp);

       // convert pixels
       pdrawLineConv(SrcP, DstP, PixCount, D3DFMT_A8R8G8B8, TextureFormat);
      end; // for k:= 0 to PHeight - 1 do

     // increment source & destination images
     Inc(SrcImg);
     Inc(DstImg);
    end;

   // unlock the texture
   Result:= Unlock(i);
   if (Result <> errNone) then
    begin
     Bitmap.Free();
     Exit;
    end;
  end;

 // release the bitmap memory
 Bitmap.Free();
 Result:= errNone;
end;

//---------------------------------------------------------------------------
function TAGFImage.imgProcess(func: TPixelFunc; Tag: Integer): Integer;
var
 LRect: TD3DLocked_Rect;
 SrcPtr: Pointer;
 Color: Cardinal;
 k, j, i, Bpp: Integer;
begin
 Result:= errNone;
 if (not FInitialized) then
  begin
   Result:= errInvalidCall;
   Exit;
  end;

 Bpp:= BytesPerPixel;

 for k:= 0 to Length(Textures) - 1 do
  begin
   if (Lock(k, LRect) <> errNone) then Exit;

   for j:= 0 to THeight - 1 do
    begin
     SrcPtr:= Pointer(Integer(LRect.pBits) + (LRect.Pitch * j));
     for i:= 0 to TWidth - 1 do
      begin
       // read source pixel
       pdrawFormatConv(SrcPtr, @Color, FFormat, D3DFMT_A8R8G8B8);

       Color:= func(Color, Tag);

       // write the altered pixel
       pdrawFormatConv(@Color, SrcPtr, D3DFMT_A8R8G8B8, FFormat);

       Inc(Integer(SrcPtr), Bpp);
      end;
    end;

   if (Unlock(k) <> errNone) then Exit;
  end;
end;

//---------------------------------------------------------------------------
function TAGFImage.negColorFunc(Color: Cardinal; Tag: Integer): Cardinal;
var
 t: Cardinal;
begin
 case Tag of
  // negative color
  1: Result:= ($FF - (Color and $FF)) or (($FF - ((Color shr 8) and $FF)) shl 8) or
   (($FF - ((Color shr 16) and $FF)) shl 16) or (Color and $FF000000);
  // negative alpha
  2: Result:= (Color and $FFFFFF) or ($FF000000 - (Color and $FF000000));
  // negative color + alpha
  3: Result:= ($FF - (Color and $FF)) or (($FF - ((Color shr 8) and $FF)) shl 8) or
   (($FF - ((Color shr 16) and $FF)) shl 16) or (($FF - ((Color shr 24) and $FF)) shl 24);
  // grayscale 
  4: begin
      t:= pdrawConv24toGray(Color);
      Result:= t + (t shl 8) + (t shl 16) + (Color and $FF000000);
     end;
  else Result:= Color;
 end;
end;

//---------------------------------------------------------------------------
function TAGFImage.imgNegative(Color, Alpha: Boolean): Integer;
var
 Tag: Integer;
begin
 Tag:= 0;
 if (Color) then
  Tag:= Tag or 1;
 if (Alpha) then
  Tag:= Tag or 2;

 if (Tag <> 0) then
  Result:= imgProcess(negColorFunc, Tag)
   else Result:= errInvalidCall;                    
end;

//---------------------------------------------------------------------------
function TAGFImage.imgGrayscale(): Integer;
begin
 Result:= imgProcess(negColorFunc, 4);
end;

//---------------------------------------------------------------------------
function TAGFImage.imgInvert(Mirror, Flip: Boolean): Integer;
var
 LRect: TD3DLocked_Rect;
 SrcPtr, DestPtr: Pointer;
 k, j, i, Bpp, MaxW, MaxH, ForW, ForH: Integer;
 Color: Cardinal;
begin
 Result:= errNone;
 if (not FInitialized) then
  begin
   Result:= errInvalidCall;
   Exit;
  end;

 Bpp:= BytesPerPixel;
 MaxW:= ImgPerW * pWidth;
 MaxH:= ImgPerH * pHeight;
 ForW:= MaxW;
 ForH:= MaxH;
 if (Mirror) then
  ForW:= ForW div 2 else
   if (Flip) then
    ForH:= ForH div 2; 

 for k:= 0 to Length(Textures) - 1 do
  begin
   if (Lock(k, LRect) <> errNone) then Exit;

   for j:= 0 to ForH - 1 do
    begin
     for i:= 0 to ForW - 1 do
      begin
       SrcPtr:= Pointer(Integer(LRect.pBits) + (LRect.Pitch * j) + (i * Bpp));
       DestPtr:= SrcPtr;
       if (Mirror) then
        begin
         if (Flip) then
          begin
           DestPtr:= Pointer(Integer(LRect.pBits) + (LRect.Pitch * ((MaxH - j) - 1)) + (((MaxW - i) - 1) * Bpp));
          end else DestPtr:= Pointer(Integer(LRect.pBits) + (LRect.Pitch * j) + (((MaxW - i) - 1) * Bpp));
        end else
        begin
         if (Flip) then
          DestPtr:= Pointer(Integer(LRect.pBits) + (LRect.Pitch * ((MaxH - j) - 1)) + (i * Bpp));
        end;

       // move pixel data
       Move(SrcPtr^, Color, Bpp);
       Move(DestPtr^, SrcPtr^, Bpp);
       Move(Color, DestPtr^, Bpp);
      end;
    end;

   if (Unlock(k) <> errNone) then Exit;
  end;
end;

//---------------------------------------------------------------------------
end.

